import numpy as np

class FileHeader:
    def __init__(self, buffer, offset=0):
        self.offset = offset
        self.file_cookie = np.frombuffer(buffer, dtype='S2', count=1, offset=self.offset).astype(str)[0]
        self.offset += 2
        self.file_version = np.frombuffer(buffer, dtype='S2', count=1, offset=self.offset).astype(str)[0]
        self.offset += 2
        self.file_size = np.frombuffer(buffer, dtype='int32', count=1, offset=self.offset).astype(np.int32)[0]
        self.offset += 4
        self.waveform_count = np.frombuffer(buffer, dtype='int32', count=1, offset=self.offset).astype(np.int32)[0]
        self.offset += 4



import numpy as np
from .manifoldutils import NewtonsMethodObject


class Cell:
    """

    """

    def __init__(self, gain, touch):
        """
        Cell object constructor.

        :param gain:
        :type gain: numpy.ndarray
        :param touch:
        :type touch: numpy.ndarray
        """

        # Check that gain is of type list or type numpy.ndarray
        if not isinstance(gain, (list, np.ndarray)):
            raise TypeError(f"Gain ({type(gain)}) must be of type list or type numpy.ndarray")

        # Check that touch is of type list or type numpy.ndarray
        if not isinstance(touch, (list, np.ndarray)):
            raise TypeError(f"Touch ({type(touch)}) must be of type list or type numpy.ndarray")

        # Convert a gain list to a numpy.ndarray
        if isinstance(gain, list):
            gain = np.array(gain)

        # Convert a touch list to a numpy.ndarray
        if isinstance(touch, list):
            touch = np.array(touch)

        # Assert that gain and touch have an equal number of dimensions except for the last dimension
        assert gain.shape[:-1] == touch.shape[:-1], f"Gain and touch must have the same dimensions up to last: " \
                                                    f"{gain.shape[:-1]}, {touch.shape[:-1]}"

        # Check that touch is of type numpy.ndarray
        assert gain.shape[-1] >= len(gain.shape[:-1]), f"Last gain dimension must have equal or more elements than " \
                                                       f"previous dimensions: "

        # Check that touch is of type numpy.ndarray
        assert touch.shape[-1] >= len(touch.shape[:-1]), f"Touch must have "

        #
        assert all([i == 2 for i in gain.shape[:-1]]), f"Unsupported cell shape: {gain.shape[:-1]}."  # Assert that
        # the first gain dimensions are equal to 2
        assert all([i == 2 for i in touch.shape[
                                    :-1]]), f"Unsupported cell shape: {touch.shape[:-1]}."  # Assert that the first
        # touch dimensions are equal to 2

        self.__gain = gain
        self.__touch = touch

    @property
    def gain(self):
        """

        """
        return self.__gain

    @property
    def touch(self):
        """

        """
        return self.__touch

    @property
    def gain_centroid(self):
        """
        Return the gain centroid computed at the midpoint of each cell axis.

        """

        return self.regress(self.gain, 0.5 * np.ones((len(self.gain.shape[:-1])), ))

    @property
    def touch_centroid(self):
        """
        Return the touch centroid computed at the midpoint of each cell axis.

        """

        return self.regress(self.touch, 0.5 * np.ones((len(self.touch.shape[:-1])), ))

    @classmethod
    def regress(cls, vertices, coords):
        """
        Regress a coordinate vector based on the given vertices.

        :param vec:
        :type vec: numpy.ndarray
        """

        # Convert singular inputs to lists
        if isinstance(coords, (int, float)):
            coords = [coords]

        # Convert regular lists to numpy arrays
        if isinstance(coords, list):
            coords = np.array(coords)

        # If the numpy array is a single dimension,
        if len(coords.shape) == 1:
            coords = coords[np.newaxis, :]

        result = np.zeros((coords.shape[0], vertices.shape[-1]))

        assert coords.shape[-1] == len(vertices.shape[
                                       :-1]), f"Coordinate vector length must match gain shape : {coords.shape[-1]}" \
                                              f" : {len(vertices.shape[:-1])}."

        for i, coord in enumerate(coords):
            tmp_vertices = vertices
            for j in reversed(range(len(coord))):  # Compute the regression for each axis
                tmp_vertices = eval(
                    f"coord[{j}]*(tmp_vertices[{':, ' * j}1, :] - tmp_vertices[{':, ' * j}0, :]) + "
                    f"tmp_vertices[{':, ' * j}0, :]")

            result[i, :] = tmp_vertices

        return result.ravel() if result.shape[0] == 1 else result

    @classmethod
    def jacobian(cls, vertices, coords):
        """
        Compute the linear Jacobian at the given vector coordinate location using the given vertices.

        """

        # This function typically shouldn't be used for multidimensional vector inputs

        # Convert singular values to lists
        if isinstance(coords, (int, float)):
            coords = [coords]

        # Assert that vec is array-like
        if not isinstance(coords, (list, np.ndarray)):
            raise TypeError(f"Vector ({type(coords)}) must be of type list or type numpy.ndarray")

        # Make sure that a NumPy array is single-dimensional
        # if isinstance(vec, np.ndarray):
        #    pass

        # Reshape vec if it is 1-dimensional
        # if len(vec.shape) == 1:
        #    vec = vec[np.newaxis, :]

        # Assert that the number of vertices dimensions to n-1 are equal to the vector length
        # assert len(vertices.shape[:-1]) == vec.shape[-1], f"Vector length must match vertices shape."

        J = np.zeros((vertices.shape[-1], 0))  # Construct the jacobian matrix
        dims = np.arange(0, len(coords), dtype=int)

        for i in range(len(coords)):
            tmp_coords = np.roll(coords, -i)  # Roll the input vector
            tmp_dims = np.roll(dims, -i)  # Roll the dimension vector
            tmp_vertices = vertices.transpose(tuple(tmp_dims) + (len(coords),))

            for i in reversed(range(1, len(coords))):
                tmp_vertices = eval(
                    f"tmp_coords[{i}]*(tmp_vertices[{':, ' * i}1, :] - tmp_vertices[{':, ' * i}0, :]) + "
                    f"tmp_vertices[{':, ' * i}0, :]")

            J = np.concatenate(
                (
                    J,
                    (tmp_vertices[1, :] - tmp_vertices[0, :])[:, np.newaxis]
                ),
                axis=1
            )

        return J

    @classmethod
    def newtons_method(cls, cell, vertices, goal, guess_vec=None, tol=1e-6, max_iter=20):
        """
        Newton's Method

        """

        # assert that the goal length must match the last vertices dimension
        assert len(goal) == vertices.shape[-1], "Goal length must match last vertices dimension."

        if guess_vec is None:
            guess_vec = 0.5 * np.ones((len(vertices.shape[:-1]),))

        I = np.identity(len(guess_vec))  # Construct an identity matrix of the length of the guess vector
        guess_vec_history = np.zeros((max_iter + 1, len(guess_vec)))
        guess_vec_history[0, :] = guess_vec

        for i in range(max_iter):
            J = cls.jacobian(vertices, guess_vec_history[i, :])
            Jinv = np.linalg.solve(J.T @ J, I)  # J.T @ J to account for non-square Jacobian matrices
            inc = (Jinv @ (J.T @ (cls.regress(vertices, guess_vec_history[i, :]) - goal)[:,
                                 np.newaxis])).ravel()  # Solve the incremental change in the guess vector

            # print(inc)
            # print(guess_vec_history[i, :])

            guess_vec_history[i + 1, :] = guess_vec_history[i, :] - inc
            # if np.all(inc < tol):
            #    break

        return NewtonsMethodObject(cell, guess_vec_history[:i + 1, :], vertices, goal, guess_vec, tol, max_iter)
        # return guess_vec_states[i+1, :], guess_vec_history[:i+1, :]




class WaveformHeader:
    def __init__(self, buffer, offset=0):
        self.offset = offset
        self.header_size = np.frombuffer(buffer, dtype='int32', count=1, offset=self.offset).astype(np.int32)[0]
        self.offset += 4
        self.waveform_type = np.frombuffer(buffer, dtype='int32', count=1, offset=self.offset).astype(np.int32)[0]
        self.offset += 4
        self.waveform_buffers = np.frombuffer(buffer, dtype='int32', count=1, offset=self.offset).astype(np.int32)[0]
        self.offset += 4
        self.points = np.frombuffer(buffer, dtype='int32', count=1, offset=self.offset).astype(np.int32)[0]
        self.offset += 4
        self.count = np.frombuffer(buffer, dtype='int32', count=1, offset=self.offset).astype(np.int32)[0]
        self.offset += 4
        self.x_display_range = np.frombuffer(buffer, dtype='float32', count=1, offset=self.offset).astype(np.float32)[0]
        self.offset += 4
        self.x_display_origin = np.frombuffer(buffer, dtype='float64', count=1, offset=self.offset).astype(np.float64)[0]
        self.offset += 8
        self.x_increment = np.frombuffer(buffer, dtype='float64', count=1, offset=self.offset).astype(np.float64)[0]
        self.offset += 8
        self.x_origin = np.frombuffer(buffer, dtype='float64', count=1, offset=self.offset).astype(np.float64)[0]
        self.offset += 8
        self.x_units = np.frombuffer(buffer, dtype='int32', count=1, offset=self.offset).astype(np.int32)[0]
        self.offset += 4
        self.y_units = np.frombuffer(buffer, dtype='int32', count=1, offset=self.offset).astype(np.int32)[0]
        self.offset += 4
        self.date_string = np.frombuffer(buffer, dtype='S16', count=1, offset=self.offset).astype(str)[0]
        self.offset += 16
        self.time_string = np.frombuffer(buffer, dtype='S16', count=1, offset=self.offset).astype(str)[0]
        self.offset += 16
        self.frame_string = np.frombuffer(buffer, dtype='S24', count=1, offset=self.offset).astype(str)[0]
        self.offset += 24
        self.waveform_string = np.frombuffer(buffer, dtype='S16', count=1, offset=self.offset).astype(str)[0]
        self.offset += 16
        self.time_tag = np.frombuffer(buffer, dtype='float64', count=1, offset=self.offset).astype(np.float64)[0]
        self.offset += 8
        self.segment_index = np.frombuffer(buffer, dtype='uint32', count=1, offset=self.offset).astype(np.uint32)[0]
        self.offset += 4
        self.offset = self.header_size + offset

        self.time = np.linspace(0, self.x_display_range, self.points, endpoint=False) - self.x_origin



import time

import serial
import socket
import os


class Connect:
    """
    The Connect class makes it easier to connect to the 2, 4 or 5 connection CTS to get live data.
    """

    def __init__(self, connections=2, data_method=None, com=None, port=None, exe_location=None):
        """
        Arguments:
        - connections (int req): States if you are connecting to either a 2, 4 or 5 connection CTS.
        - data_method (str req): States what data retrieval method is being used. Either 'com',
        'socket', or 'exe'.
        - com (str): Value of COM point to connect to (ex. 'COM3'). Required if data_method is 'com'.
        - port (int): Value of socket port to connect to (ex. 8888). Required if data_method is 'socket'.
        - exe_location (str): Value of location to visual_backend_v1.exe
        (ex. 'c:/Users/user/Desktop/visual_backend_v1.exe'). Required if data_method is 'exe'.
        """

        self.ser = None
        self.mainSocket = None

        if connections == 2 or connections == 4 or connections == 5:
            self.connections = connections
        else:
            raise Exception('Value for connections must be either 2, 4 or 5.')

        data_method = data_method.lower().strip()

        if data_method == 'com' or data_method == 'socket' or data_method == 'exe':
            self.data_method = data_method
            self.com = com
            self.port = port
            self.exe_location = exe_location
        else:
            raise Exception('Value for data_method must be com, socket or exe.')

        if self.data_method == 'com':
            if self.com is None or not isinstance(self.com, str):
                raise Exception('When data_method value is com... com must have a string value.')
        elif self.data_method == 'socket':
            if self.port is None or not isinstance(self.port, int):
                raise Exception('When data_method value is socket... port must have a integer value.')
        elif self.data_method == 'exe':
            if self.exe_location is None or not isinstance(self.exe_location, str):
                raise Exception('When data_method value is exe... exe_location must have a string value.')

    def init(self, open_exe=False):
        """
        Sets up the Connect class and returns the class. A required step before running any functions.
        """

        if open_exe:
            self.open(exe_location=self.exe_location)
            time.sleep(5)

        if self.data_method == 'com':
            port = str(self.com)
            if os.name == 'nt':
                port = "\\\\.\\" + str(self.com)
            self.ser = serial.Serial(port, 4608000, timeout=0)
        elif self.data_method == 'socket':
            self.mainSocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            remote_ip = socket.gethostbyname(socket.gethostname())
            self.mainSocket.connect((remote_ip, self.port))
        elif self.data_method == 'exe':
            return Exception("The data_method exe has not been implemented yet")

        return self

    def open(self, exe_location, params=None):
        if self.data_method == 'com':
            raise Exception("This function open is not available for this data_method")
        elif self.data_method == 'socket':
            os.system(exe_location)
        elif self.data_method == 'exe':
            return Exception("The data_method exe has not been implemented yet")

        return self

    def read_line(self, do_check=True):
        """
        Returns string of the latest data from the CTS. Returns None if no data is there.

        Arguments:
        - do_check (bool): Default is True. This value indicates if data_method is 'com' that it will check
        if there is data to return.
        """

        if self.data_method == 'com':
            if (do_check and self.size() > 30) or not do_check:
                return self.ser.readline()
            else:
                return None
        elif self.data_method == 'socket':
            return self.mainSocket.recv(1024)
        elif self.data_method == 'exe':
            return Exception("The data_method exe has not been implemented yet")

    def get_cords(self, do_check=True):
        """
        Returns array of the latest data from the CTS. Returns empty array if no data is there.

        Arguments:
        - do_check (bool): Default is True. This value indicates if data_method is 'com' that it will check
        if there is data to return.
        """

        read = self.read_line(do_check)
        if read is not None:
            tmp = read.decode().split(',')

            if len(tmp) == self.connections:
                return list(map(float, tmp))
            else:
                return []
        else:
            return []

    def size(self):
        """
        Only if data_method is 'com' return the bytes (int) waiting in the queue.
        """

        if self.data_method == 'com':
            return self.ser.inWaiting()
        else:
            raise Exception("This function size is not available for this data_method")

    def flush(self):
        """
        Only if data_method is 'com' flush the data waiting in queue.
        """

        if self.data_method == 'com':
            self.ser.flush()
        else:
            raise Exception("This function flush is not available for this data_method")

    def end(self):
        """
        End process of receiving data for cts. If any functions are run after this they will throw
        errors.
        """

        if self.data_method == 'com':
            self.ser.close()
        elif self.data_method == 'socket':
            self.mainSocket.close()
        elif self.data_method == 'exe':
            return Exception("The data_method exe has not been implemented yet")
